# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2013-2015, Michigan State University.
# Copyright (C) 2015-2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=missing-docstring,no-member,invalid-name,unused-variable
from __future__ import print_function
from __future__ import absolute_import

import khmer
from . import khmer_tst_utils as utils


def pretty_compare(a, b):
    print(len(a), len(b))

    line1 = []
    line2 = []
    line3 = []
    for (x, y) in zip(a, b):
        line1.append(x)
        line2.append(y)
        if x == y:
            line3.append('|')
        else:
            line3.append('x')

    for i in range(0, len(line1), 60):
        print("".join(line1[i:i + 60]))
        print("".join(line3[i:i + 60]))
        print("".join(line2[i:i + 60]))


def eq_(v1, v2):
    assert len(v1)
    if v1 != v2:
        pretty_compare(v1, v2)
    assert v1 == v2, (v1, v2)


def neq_(v1, v2):
    assert len(v1)
    if v1 == v2:
        pretty_compare(v1, v2)
    assert v1 != v2, (v1, v2)


def test_graph_attribute():
    ch = khmer.Countgraph(10, 1048576, 1)
    aligner = khmer.ReadAligner(ch, 0, 0)
    assert aligner.graph is ch


def test_align_nothing():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "ACCAAGGCTCGAGATTTACC"

    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    score, graphAlign, readAlign, trunc = aligner.align(read)

    print(score, graphAlign, readAlign)

    assert trunc
    assert len(graphAlign) == 0
    assert len(readAlign) == 0


def test_alignnocov():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "ACCTAGGTTCGACATGTACC"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    ch.consume("ACCTAGGTTCGACATGTACC")
    _, graphAlign, readAlign, trunc = aligner.align(read)

    # should be the same
    eq_(readAlign, 'ACCTAGGTTCGACATGTACC')
    eq_(graphAlign, 'ACCTAGGTTCGACATGTACC')
    assert not trunc


def test_align_middle():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "TCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    ch.consume(read)
    _, graphAlign, readAlign, trunc = aligner.align(read)

    # should be the same
    eq_(readAlign, read)
    eq_(graphAlign, read)
    assert not trunc


def test_align_middle_trunc():
    return  # @CTB

    ch = khmer.Countgraph(10, 1048576, 1)
    read = "TCGACAAGTCCTTGACAGATGGGGGG"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")

    # omit suffix from graph
    ch.consume(read[:-5])
    _, graphAlign, readAlign, trunc = aligner.align(read)

    # should not be the same...
    neq_(readAlign, read)
    neq_(graphAlign, read)

    eq_(readAlign, read[:-5])
    eq_(graphAlign, read[:-5])

    # ...but truncated
    assert trunc


def test_align_middle_trunc_2():
    return  # @CTB

    ch = khmer.Countgraph(10, 1048576, 1)
    read = "GGGGGGGGGGGGTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AAAAAAAAAAAATCGACAAGTCCTTGACAGAT")

    # omit prefix from graph
    ch.consume(read[12:])
    _, graphAlign, readAlign, trunc = aligner.align(read)

    # here, the alignment must start not at the beginning
    print(readAlign)
    print(graphAlign)

    eq_(readAlign, read[12:])
    eq_(graphAlign, read[12:])

    # ...but truncated
    assert trunc


def test_align_fwd_nothing():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "ACCAAGGCTCGAGATTTACC"

    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    score, graphAlign, readAlign, trunc, _ = aligner.align_forward(read)

    print(score, graphAlign, readAlign)

    assert trunc
    assert len(graphAlign) == 0
    assert len(readAlign) == 0


def test_align_fwd_nocov():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "ACCTAGGTTCGACATGTACC"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    ch.consume("ACCTAGGTTCGACATGTACC")
    _, graphAlign, readAlign, trunc, _ = aligner.align_forward(read)

    # should be the same
    eq_(readAlign, 'ACCTAGGTTCGACATGTACC')
    eq_(graphAlign, 'ACCTAGGTTCGACATGTACC')
    assert not trunc


def test_align_fwd_middle():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "TCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    ch.consume(read)
    score, graphAlign, readAlign, trunc, _ = aligner.align_forward(read)

    # should be the same
    eq_(readAlign, read)
    eq_(graphAlign, read)
    assert not trunc


def test_align_fwd_middle_trunc():
    return  # @CTB
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "TCGACAAGTCCTTGACAGATGGGGGG"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for i in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")

    # omit suffix from graph
    ch.consume(read[:-5])
    score, graphAlign, readAlign, trunc, _ = aligner.align_forward(read)

    # should not be the same...
    neq_(readAlign, read)
    neq_(graphAlign, read)

    eq_(readAlign, read[:-5])
    eq_(graphAlign, read[:-5])

    # ...but truncated
    assert trunc


def test_align_fwd_middle_trunc_2():
    ch = khmer.Countgraph(10, 1048576, 1)
    read = "GGGGGGGGGGGGTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for _ in range(20):
        ch.consume("AAAAAAAAAAAATCGACAAGTCCTTGACAGAT")

    # omit prefix from graph
    ch.consume(read[12:])
    _, graphAlign, readAlign, trunc, _ = aligner.align_forward(read)

    # this will fail, because align_forward chooses the first kmer as the
    # seed.
    assert not readAlign
    assert not graphAlign
    assert trunc


def test_align_fwd_covs_1():
    K = 10

    ch = khmer.Countgraph(K, 1048576, 1)
    read = "GTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for i in range(19):
        ch.consume(read)

    ch.consume("CTCGACAAGTCCTTGACAGAT")
    #           ^
    score, g, r, is_t, covs = aligner.align_forward(read)

    for start in range(0, len(read) - K + 1):
        print(ch.get(read[start:start + K]), end=' ')
    print('')

    assert len(covs) == len(read)
    assert covs[0] == 19
    assert min(covs[1:-K]) == 20, covs
    assert max(covs) == 20, covs


def test_align_fwd_covs_2():
    K = 10

    ch = khmer.Countgraph(K, 1048576, 1)
    read = "GTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for i in range(19):
        ch.consume(read)

    ch.consume("GACGACAAGTCCTTGACAGAT")
    #            ^
    score, g, r, is_t, covs = aligner.align_forward(read)

    print(covs, g)
    for start in range(0, len(read) - K + 1):
        print(ch.get(read[start:start + K]), end=' ')
    print('')

    assert len(covs) == len(read)
    assert covs[0] == 19
    assert covs[1] == 19
    assert min(covs[2:-K]) == 20, covs
    assert max(covs) == 20, covs


def test_align_fwd_covs_3():
    K = 10

    ch = khmer.Countgraph(K, 1048576, 1)
    read = "GTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for i in range(19):
        ch.consume(read)

    ch.consume("GTAGACAAGTCCTTGACAGAT")
    #             ^
    score, g, r, is_t, covs = aligner.align_forward(read)

    print(covs, g)
    for start in range(0, len(read) - K + 1):
        print(ch.get(read[start:start + K]), end=' ')
    print('')

    assert len(covs) == len(read)
    assert covs[0] == 19
    assert covs[1] == 19
    assert covs[2] == 19
    assert min(covs[3:-K]) == 20, covs
    assert max(covs) == 20, covs


def test_align_fwd_covs_4():
    K = 10

    ch = khmer.Countgraph(K, 1048576, 1)
    read = "GTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for i in range(19):
        ch.consume(read)

    ch.consume("GTCGACAAGTCCTTGACAGAG")
    #                               ^
    score, g, r, is_t, covs = aligner.align_forward(read)

    print(covs, g)
    for start in range(0, len(read) - K + 1):
        print(ch.get(read[start:start + K]), end=' ')
    print('')

    assert len(covs) == len(read)
    assert covs[-K] == 19
    assert min(covs[:-K]) == 20, covs
    assert max(covs) == 20, covs


def test_align_fwd_covs_5():
    K = 10

    ch = khmer.Countgraph(K, 1048576, 1)
    read = "GTCGACAAGTCCTTGACAGAT"
    aligner = khmer.ReadAligner(ch, 0, 0)
    for i in range(19):
        ch.consume(read)

    ch.consume("GTCGACAAGTCCTTGACAGCT")
    #                              ^
    score, g, r, is_t, covs = aligner.align_forward(read)

    print(covs, g)
    for start in range(0, len(read) - K + 1):
        print(ch.get(read[start:start + K]), end=' ')
    print('')

    assert len(covs) == len(read)
    assert covs[-K] == 19
    assert covs[-K - 1] == 19
    assert min(covs[:-K - 1]) == 20, covs
    assert max(covs) == 20, covs


def test_simple_readalign():
    return  # @CTB
    ch = khmer.Countgraph(10, 1048576, 1)
    aligner = khmer.ReadAligner(ch, 2, 0)
    for i in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACATGTCCTTGACAGAT")
    read = "ACCTAGGTTCGACAAGTACC"
    #                      ^^            ^  ^
    ch.consume("GCTTTTAAAAAGGTTCGACAAAGGCCCGGG")
    # CCCGGGCCTTTGTCGAACCTTTTTAAAAGC

    score, graphAlign, readAlign, trunc = aligner.align(read)

#                        AGCTAGGTTCGACAAGT CCT
#                        ACCTAGGTTCGACAAGTaCC
#                        --CTAGGTTCGACATGT-CC
    eq_(graphAlign, 'AGCTAGGTTCGACATGTCCT')
    eq_(readAlign, 'ACCTAGGTTCGACAAGTACC')


def test_readalign():
    return  # @CTB
    ch = khmer.Countgraph(10, 1048576, 1)
    aligner = khmer.ReadAligner(ch, 1, 0)
    for i in range(20):
        ch.consume("AGAGGGAAAGCTAGGTTCGACAAGTCCTTGACAGAT")
    read = "ACCTAGGTTCGACATGTACC"
    #                      ^^            ^  ^

    ch.consume("GCTTTTAAAAAGGTTCGACAAAGGCCCGGG")

    score, graphAlign, readAlign, _ = aligner.align(read)

    eq_(readAlign, 'ACCTAGGTTCGACATGTACC')
    eq_(graphAlign, 'AGCTAGGTTCGACAAGTCCT')


ht_seqs = ["TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAACTGG"
           "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTGTTGCAATCTTAACAAC"
           "CTCTTTAC",
           "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAACTGG"
           "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTGTTGCAATCTTAACAAC"
           "CTCTTTAC",
           "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAACTGG"
           "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
           "CTCTTTAC",
           "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAACTGG"
           "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGAAAAATGTCATCCTGTATTGCAATCTTAACAAC"
           "CTCTTTAC"]

queries = [
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAA"
               "CTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTGTTGCAATC"
               "TTAACAACCTCTTTAC",
        "score": 274.76338282696173,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCG"
                     "CTTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCC"
                     "TGTGTTGCAATCTTAACAACCTCTTTAC",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGC"
                    "TTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTG"
                    "TGTTGCAATCTTAACAACCTCTTTAC",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAA"
               "CTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATC"
               "TAACAACCTCTTTAC",
        "score": 274.76338282696173,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCG"
                     "CTTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCC"
                     "TGTATTGCAATCTTAACAACCTCTTTAC",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGC"
                    "TTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTG"
                    "TATTGCAATCTTAACAACCTCTTTAC",
        "truncated": False
    },
    {
        "seq": "TAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAAC"
               "TGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTGTTGCAATCT"
               "TAACAACCTCTTTAC",
        "score": 272.841515695261,
        "graph_aln": "TAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGC"
                     "TTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCT"
                     "GTGTTGCAATCTTAACAACCTCTTTAC",
        "read_aln": "TAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCT"
                    "TTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGT"
                    "GTTGCAATCTTAACAACCTCTTTAC",
        "truncated": False
    },
    {
        "seq": "TAAATGCGCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAAC"
               "TGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTGTTGCAATCT"
               "TAACAACCTCTTTAC",
        "score": 268.2640868672253,
        "graph_aln": "TAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGC"
                     "TTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCT"
                     "GTGTTGCAATCTTAACAACCTCTTTAC",
        "read_aln": "TAAATGCGCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCT"
                    "TTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGT"
                    "GTTGCAATCTTAACAACCTCTTTAC",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAA",
        "score": 97.37145206396536,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAA",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTAGATGTTTGATTATCAA",
        "score": 92.79402323592961,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAA",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTAGATGTTTGATTATCAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTATTGATTATCAA",
        "score": 84.74620322710143,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGT-TTGATTATCAA",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTaTTGATTATCAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATTGTTTGATTATCAA",
        "score": 82.2182409986759,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATaTGTTTGATTATCAA",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTAT-TGTTTGATTATCAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTATTGATTATCAA",
        "score": 84.74620322710143,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGT-TTGATTATCAA",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTaTTGATTATCAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTATAGATTATCAA",
        "score": 80.1687743990657,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGT-TTGATTATCAA",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTaTAGATTATCAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATAATTTTGCCGCTTTAAC"
               "TGGGTCTAGTTTCTACTGCAAACTTTCCACCAACTAGTTTTTCTGCATCCTTTGTTGCAATC"
               "TTAACAACCTCTTTAC",
        "score": 237.81111469018322,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATcAATTTTGCCG"
                     "CTTTAACTGGGTCT-GTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATC"
                     "CTGTGTTGCAATCTTAACAACCTCTTTAC",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTAT-AATTTTGCCGC"
                    "TTTAACTGGGTCTaGTTTCTACTGCAAACTTTCCACCAACTAGTTTTTCTGCATCCT"
                    "TTGTTGCAATCTTAACAACCTCTTTAC",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGAAAATAATTAAAAAAAAAAAAA"
               "AAAAAAAAAAAAAAAAAAAAAAAAAA",
        "score": 5.331560863368736,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCG"
                     "CTTTAACTGGGTCTGTTTCTACTGCAAACTTT",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGAAAATAATTAAAAAAAA"
                    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        "truncated": False
    },
    {
        "seq": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAA"
               "CTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGAAAAATGTCATCCTGTATTGCAATC"
               "TTAACAACCTCTTTAC",
        "score": 274.76338282696173,
        "graph_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCG"
                     "CTTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGAAAAATGTCATCC"
                     "TGTATTGCAATCTTAACAACCTCTTTAC",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGC"
                    "TTTAACTGGGTCTGTTTCTACTGCAAACTTTCCACCAACAAGAAAAATGTCATCCTG"
                    "TATTGCAATCTTAACAACCTCTTTAC",
        "truncated": False
    },
    {  # the motif of 32 bases are identical match to HT seqs, the rest are
        # random. TTAAATGCCCAATTTTTCCCTCTTTTCTTCTAT" is the from HT seqs
        "seq":
        "ACAAGGCCATTTGTTCGCATTCTGAAGCCGGCTTCCACCATGGTACTGGGAAACTGTCGGAATATTAAA"
        "TGCCCAATTTTTCCCTCTTTTCTTCTATCCGCAGTATGGACACTGTTTTCCTGAATTTCATTGACAGTT"
        "TAATTTACTGCGGTCACGCGGAACT",
        "score": 68.17022311739733,
        "graph_aln":
        "ACAAGGCCATTTGTTCGCATTCTGAAGCCGGCTTCCACCATGGTACTGGGAAACTGTCGGAATATTAAA"
        "TGCCCAATTTTTCCCTCTTTTCTTCTATCCGCAGTATGGACACTGTTTTCCTGAATTTCATTGACAGTT"
        "TAATTTACTGCGGTCACGCGGAACT",
        "read_aln": "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTAT",
        "truncated": True,
        "description": "truncated-alignment-bc-missing-kmers"
    },
    {   # Testing for min distance between correctable SNPs
        # 1st SNP is at position 2+K from beginning, 2nd SNP at position 2+K+K
        "seq":
        "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATACGTTTGATTATCAATTTTGCCGCTTTAACTGG"
        "ATCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
        "CTCTTTAC",
        "score": 265.608525171,
        "graph_aln":
        "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAACTGG"
        "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
        "CTCTTTAC",
        "read_aln":
        "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATACGTTTGATTATCAATTTTGCCGCTTTAACTGG"
        "ATCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
        "CTCTTTAC",
        "truncated": False,
        "description": "2 SNPs, one K apart",
    },
    {   # Testing for min distance between correctable SNPs
        # 1st SNP is at position 2+K from beginning, 2nd SNP at position
        # 2+K+K-1
        "seq":
        "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATACCTTTGATTATCAATTTTGCCGCTTTAACTGG"
        "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
        "CTCTTTAC",
        "score": 265.608525171,
        "graph_aln":
        "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATATGTTTGATTATCAATTTTGCCGCTTTAACTGG"
        "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
        "CTCTTTAC",
        "read_aln":
        "TTAAATGCCCAATTTTTCCCTCTTTTCTTCTATACGTTTGATTATCAATTTTGCCGCTTTAACTAG"
        "GTCTGTTTCTACTGCAAACTTTCCACCAACAAGTTTTTCTGCATCCTGTATTGCAATCTTAACAAC"
        "CTCTTTAC",
        "truncated": False,
        "description": "2 SNPs, K-2 apart",
    }

]


def check_query(aligner, query):
    score, graphAlign, readAlign, trunc = aligner.align(query["seq"])
    print(query["seq"])
    print(graphAlign, query["graph_aln"])
    print(readAlign, query["read_aln"])
    print(trunc, query["truncated"])
    print(score, query["score"])
    assert graphAlign == query["graph_aln"], "\n%r != \n%r" % \
        (graphAlign, query["graph_aln"])
    assert readAlign == query["read_aln"], "\n%r != \n%r" % \
        (readAlign, query["read_aln"])
    eq_(trunc, query["truncated"])
    if query["score"] > 0:
        assert round(score - query["score"], 7) == 0


def test_readalign_new():
    return  # @CTB
    ch = khmer.Countgraph(32, 1048576, 1)
    aligner = khmer.ReadAligner(ch, 1, 0)
    for seq in ht_seqs:
        ch.consume(seq)

    for query in queries:
        if "description" in query:
            check_query.description = query["description"]
        yield check_query, aligner, query


def test_readaligner_load():
    ct = khmer.Countgraph(32, 1048576, 1)
    parameters_json = utils.get_test_data('readaligner-default.json')
    a_aligner = khmer.ReadAligner(ct, 0, 0, filename=parameters_json)
    a_scoring_matrix = a_aligner.get_scoring_matrix()
    a_transition_probabilities = a_aligner.get_transition_probabilities()
    assert a_scoring_matrix[0] == -0.06642736173897607, a_scoring_matrix[0]
    assert a_transition_probabilities[0][0] == -0.021973842014145723, (
        a_transition_probabilities[0][0])

    for seq in ht_seqs:
        ct.consume(seq)

    for query in queries:
        a_aligner.align(query['seq'])

    b_aligner = khmer.ReadAligner(
        ct, 0, 0, transition_probabilities=a_transition_probabilities,
        scoring_matrix=a_scoring_matrix)
    b_scoring_matrix = b_aligner.get_scoring_matrix()
    b_transition_probabilities = b_aligner.get_transition_probabilities()
    assert b_scoring_matrix == a_scoring_matrix, (
        a_scoring_matrix, b_scoring_matrix)
    assert b_transition_probabilities == a_transition_probabilities, (
        a_transition_probabilities, b_transition_probabilities)



from __future__ import print_function
from __future__ import absolute_import

import gc
import itertools
import random

import khmer
from khmer._oxli.parsing import Sequence, FastxParser, SanitizedFastxParser
from khmer._oxli.parsing import BrokenPairedReader, Alphabets, check_is_pair
from khmer._oxli.parsing import check_is_right, check_is_left
from khmer.khmer_args import estimate_optimal_with_K_and_f as optimal_fp
from khmer import reverse_complement as revcomp
from khmer import reverse_hash as revhash
from . import khmer_tst_utils as utils

import pytest
import screed


def teardown():
    utils.cleanup()


@pytest.fixture
def create_fastx(tmpdir, as_str=True):
    def func(reads, fmt='fa'):
        assert fmt in ['fa', 'fq']
        fastx_fn = tmpdir.join('test.' + fmt)
        for record in reads:
            if fmt == 'fa':
                fastx_fn.write('>{0}\n{1}\n'.format(record.name,
                                                    record.sequence),
                               mode='a')
            else:
                fastx_fn.write('@{0}\n{1}\n+\n{2}\n'.format(record.name,
                                                            record.sequence,
                                                            record.quality),
                               mode='a')
        return str(fastx_fn) if as_str else fastx_fn
    return func


def sequences_eq(seqs_A, seqs_B):
    for seq1, seq2 in zip(seqs_A, seqs_B):
        if seq1 is None:
            assert seq1 is seq2
        else:
            assert seq1.name == seq2.name
            assert seq1.sequence == seq2.sequence


def test_FastxParser(create_fastx):
    expected = [Sequence('seq1/1', 'A' * 5),
                Sequence('seq1/2', 'A' * 4),
                Sequence('seq2/1', 'A' * 5),
                Sequence('seq3/1', 'A' * 3),
                Sequence('seq3/2', 'A' * 5)]
    parser = FastxParser(create_fastx(expected))
    result = list(parser)

    assert len(expected) == len(result)
    assert all((x == y) for x, y in zip(expected, result))


def test_SanitizedFastxParser_convert_Ns(create_fastx):
    '''Test that A's are converted to N's'''
    expected = [Sequence('seq1/1', 'N' * 5),
                Sequence('seq1/2', 'N' * 4)]
    parser = SanitizedFastxParser(create_fastx(expected),
                                  alphabet='DNAN_SIMPLE')
    result = list(parser)

    assert parser.n_bad == 0
    assert len(result) == 2
    assert result[0].sequence == 'A' * 5
    assert result[1].sequence == 'A' * 4


def test_SanitizedFastxParser_no_convert_Ns(create_fastx):
    expected = [Sequence('seq1/1', 'N' * 5),
                Sequence('seq1/2', 'N' * 4)]
    parser = SanitizedFastxParser(create_fastx(expected),
                                  alphabet='DNAN_SIMPLE',
                                  convert_n=False)
    result = list(parser)

    assert parser.n_bad == 0
    assert len(result) == 2
    assert result[0].sequence == 'N' * 5
    assert result[1].sequence == 'N' * 4


def test_SanitizedFastxParser_invalid(create_fastx):
    '''Test that parser detects invalid sequence'''
    expected = [Sequence('seq1/1', 'XXX'),
                Sequence('seq1/2', 'A' * 4)]
    parser = SanitizedFastxParser(create_fastx(expected))
    result = list(parser)

    assert parser.n_bad == 1
    assert len(result) == 1
    assert result[0].sequence == 'A' * 4


def test_SanitizedFastxParser_lowercase(create_fastx):
    reads = [Sequence('seq1/1', 'acgtn'),
             Sequence('seq1/2', 'AcGtN'),
             Sequence('seq1/2', 'aCgTn')]

    parser = SanitizedFastxParser(create_fastx(reads), convert_n=False)
    result = list(parser)

    assert result[0].sequence == 'ACGTN'
    assert result[1].sequence == 'ACGTN'
    assert result[2].sequence == 'ACGTN'


def test_alphabet_wrapper():
    dna_simple = Alphabets.get('DNA_SIMPLE')
    assert len(dna_simple) == 4
    for b in 'ACGT':
        assert b in dna_simple

    with pytest.raises(ValueError):
        Alphabets.get('TEST')


def gather_paired(stream, **kw):
    itr = BrokenPairedReader(stream, **kw)

    x = []
    m = 0
    num = 0
    for num, is_pair, read1, read2 in itr:
        x.append((read1.name if read1 is not None else None,
                  read2.name if read2 is not None else None))
        m += 1

    return x, num, m


class Test_BrokenPairedReader(object):
    reads = [Sequence(name='seq1/1', sequence='A' * 5),
             Sequence(name='seq1/2', sequence='A' * 4),
             Sequence(name='seq2/1', sequence='A' * 5),
             Sequence(name='seq3/1', sequence='A' * 3),
             Sequence(name='seq3/2', sequence='A' * 5)]

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testDefault(self, parser, create_fastx):
        x, n, m = gather_paired(parser(create_fastx(self.reads)),
                                min_length=1)

        expected = [('seq1/1', 'seq1/2'),
                    ('seq2/1', None),
                    ('seq3/1', 'seq3/2')]
        assert x == expected, x
        assert m == 3
        assert n == 3, n

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testMinLength(self, parser, create_fastx):
        x, n, m = gather_paired(parser(create_fastx(self.reads)),
                                min_length=3)

        expected = [('seq1/1', 'seq1/2'),
                    ('seq2/1', None),
                    ('seq3/1', 'seq3/2')]
        assert x == expected, x
        assert m == 3
        assert n == 3, n

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testMinLength_2(self, parser, create_fastx):
        x, n, m = gather_paired(parser(create_fastx(self.reads)),
                                min_length=4)

        expected = [('seq1/1', 'seq1/2'),
                    ('seq2/1', None),
                    (None, 'seq3/2')]
        assert x == expected, x
        assert m == 3
        assert n == 3, n

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testForceSingle(self, parser, create_fastx):
        x, n, m = gather_paired(parser(create_fastx(self.reads)),
                                force_single=True)

        expected = [('seq1/1', None),
                    ('seq1/2', None),
                    ('seq2/1', None),
                    ('seq3/1', None),
                    ('seq3/2', None)]
        assert x == expected, x
        assert m == 5
        assert n == 4, n

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testForceSingleAndMinLength(self, parser, create_fastx):
        x, n, m = gather_paired(parser(create_fastx(self.reads)),
                                min_length=5, force_single=True)

        expected = [('seq1/1', None),
                    ('seq2/1', None),
                    ('seq3/2', None)]
        assert x == expected, x
        assert m == 3, m
        assert n == 2, n

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testRequirePairedAndMinLength_HalfPass(self, parser, create_fastx):
        reads = [Sequence('seq1/1', 'A' * 5),
                 Sequence('seq1/2', 'A' * 4),
                 Sequence('seq3/1', 'A' * 3),
                 Sequence('seq3/2', 'A' * 5)]

        reader = BrokenPairedReader(parser(create_fastx(reads)),
                                    min_length=4, require_paired=True)

        result = []
        for n, paired, first, second in reader:
            result.append((first, second))

        assert len(result) == 1
        assert n == 0
        l, r = result[0]
        assert l == reads[0]
        assert r == reads[1]

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testRequirePairedAndMinLength_SwappedHalfPass(self, parser,
                                                      create_fastx):
        reads = [Sequence('seq1/1', 'A' * 5),
                 Sequence('seq1/2', 'A' * 4),
                 Sequence('seq3/1', 'A' * 5),
                 Sequence('seq3/2', 'A' * 3)]

        reader = BrokenPairedReader(parser(create_fastx(reads)),
                                    min_length=4, require_paired=True)

        result = []
        for n, paired, first, second in reader:
            result.append((first, second))

        assert n == 0
        assert len(result) == 1
        l, r = result[0]
        assert l == reads[0]
        assert r == reads[1]

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testRequirePairedAndMinLength_NeitherPass(self, parser, create_fastx):
        reads = [Sequence('seq1/1', 'A' * 5),
                 Sequence('seq1/2', 'A' * 4),
                 Sequence('seq3/1', 'A' * 3),
                 Sequence('seq3/2', 'A' * 3)]

        reader = BrokenPairedReader(parser(create_fastx(reads)),
                                    min_length=4, require_paired=True)

        result = []
        for n, paired, first, second in reader:
            result.append((first, second))

        assert n == 0
        assert len(result) == 1
        l, r = result[0]
        assert l == reads[0]
        assert r == reads[1]

    @pytest.mark.parametrize("parser", [FastxParser, SanitizedFastxParser])
    def testRequirePairedAndMinLength_SwappedNeitherPass(self, parser,
                                                         create_fastx):
        reads = [Sequence('seq1/1', 'A' * 3),
                 Sequence('seq1/2', 'A' * 3),
                 Sequence('seq3/1', 'A' * 5),
                 Sequence('seq3/2', 'A' * 5)]

        reader = BrokenPairedReader(parser(create_fastx(reads)),
                                    min_length=4, require_paired=True)

        result = []
        for n, paired, first, second in reader:
            result.append((first, second))

        assert n == 0
        assert len(result) == 1
        l, r = result[0]
        assert l == reads[2]
        assert r == reads[3]


def test_check_is_pair_1():
    read1 = Sequence(name='seq', quality='###', sequence='AAA')
    read2 = Sequence(name='seq2', quality='###', sequence='AAA')

    assert not check_is_pair(read1, read2)


def test_check_is_pair_2():
    read1 = Sequence(name='seq/1', quality='###', sequence='AAA')
    read2 = Sequence(name='seq/2', quality='###', sequence='AAA')

    assert check_is_pair(read1, read2)


def test_check_is_pair_3_fq():
    read1 = Sequence(name='seq 1::', quality='###', sequence='AAA')
    read2 = Sequence(name='seq 2::', quality='###', sequence='AAA')

    assert check_is_pair(read1, read2)


def test_check_is_pair_3_broken_fq_1():
    read1 = Sequence(name='seq', quality='###', sequence='AAA')
    read2 = Sequence(name='seq 2::', quality='###', sequence='AAA')

    assert not check_is_pair(read1, read2)


def test_check_is_pair_3_broken_fq_2():
    read1 = Sequence(name='seq 1::', quality='###', sequence='AAA')
    read2 = Sequence(name='seq', quality='###', sequence='AAA')

    assert not check_is_pair(read1, read2)


def test_check_is_pair_3_fa():
    read1 = Sequence(name='seq 1::', sequence='AAA')
    read2 = Sequence(name='seq 2::', sequence='AAA')

    assert check_is_pair(read1, read2)


def test_check_is_pair_4():
    read1 = Sequence(name='seq/1', quality='###', sequence='AAA')
    read2 = Sequence(name='seq/2', sequence='AAA')

    try:
        check_is_pair(read1, read2)
        assert False                    # check_is_pair should fail here.
    except ValueError:
        pass


def test_check_is_pair_4b():
    read1 = Sequence(name='seq/1', sequence='AAA')
    read2 = Sequence(name='seq/2', quality='###', sequence='AAA')

    try:
        check_is_pair(read1, read2)
        assert False                    # check_is_pair should fail here.
    except ValueError:
        pass


def test_check_is_pair_5():
    read1 = Sequence(name='seq/1', sequence='AAA')
    read2 = Sequence(name='seq/2', sequence='AAA')

    assert check_is_pair(read1, read2)


def test_check_is_pair_6():
    read1 = Sequence(name='seq1', sequence='AAA')
    read2 = Sequence(name='seq2', sequence='AAA')

    assert not check_is_pair(read1, read2)


def test_check_is_pair_7():
    read1 = Sequence(name='seq/2', sequence='AAA')
    read2 = Sequence(name='seq/1', sequence='AAA')

    assert not check_is_pair(read1, read2)


def test_check_is_right():
    assert not check_is_right('seq1/1')
    assert not check_is_right('seq1 1::N')
    assert check_is_right('seq1/2')
    assert check_is_right('seq1 2::N')

    assert not check_is_right('seq')
    assert not check_is_right('seq 2')


def test_check_is_left():
    assert check_is_left('seq1/1')
    assert check_is_left('seq1 1::N')
    assert not check_is_left('seq1/2')
    assert not check_is_left('seq1 2::N')

    assert not check_is_left('seq')
    assert not check_is_left('seq 1')

    assert check_is_left(
        '@HWI-ST412:261:d15khacxx:8:1101:3149:2157 1:N:0:ATCACG')


class Test_Sequence(object):

    name = 'Test'
    sequence = 'ACGT'
    quality = '####'
    description = 'The nucleotides'
    cleaned = 'aaaa'

    def test_init_name_and_sequence(self):
        s = Sequence(name=self.name, sequence=self.sequence)
        assert s.name == self.name
        assert s.sequence == self.sequence
        assert s.quality is None
        assert s.description is None
        assert s.cleaned_seq == self.sequence

    def test_init_name_only(self):
        s = Sequence(name=self.name)
        assert s.name is None
        assert s.sequence is None
        assert s.quality is None
        assert s.description is None
        assert s.cleaned_seq is None

    def test_init_sequence_only(self):
        s = Sequence(sequence=self.sequence)
        assert s.name is None
        assert s.sequence is None
        assert s.quality is None
        assert s.description is None
        assert s.cleaned_seq is None

    def test_init_with_cleaned_seq(self):
        s = Sequence(name=self.name, sequence=self.sequence,
                     cleaned_seq=self.cleaned)
        assert s.name == self.name
        assert s.sequence == self.sequence
        assert s.quality is None
        assert s.description is None
        assert s.cleaned_seq == self.cleaned



#! /usr/bin/env python
# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2010-2015, Michigan State University.
# Copyright (C) 2015, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=missing-docstring,no-member
from __future__ import print_function
from __future__ import absolute_import

import khmer
import sys
import screed
from screed.fasta import fasta_iter


def main():
    filename = sys.argv[1]
    K = int(sys.argv[2])  # size of kmer
    HT_SIZE = int(sys.argv[3])  # size of hashtable
    N_HT = int(sys.argv[4])  # number of hashtables

    ht = khmer.Nodegraph(K, HT_SIZE, N_HT)

    n_unique = 0
    for n, record in enumerate(fasta_iter(open(filename))):
        sequence = record['sequence']
        seq_len = len(sequence)
        for n in range(0, seq_len + 1 - K):
            kmer = sequence[n:n + K]
            if (not ht.get(kmer)):
                n_unique += 1
            ht.count(kmer)

    print(n_unique)
    print(ht.n_occupied())
    print(ht.n_unique_kmers())


if __name__ == '__main__':
    main()



# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2010-2015, Michigan State University.
# Copyright (C) 2015-2016, The Regents of the University of California.
# Copyright (C) 2016, Google, Inc
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=missing-docstring

from __future__ import print_function
import tempfile
import os
import shutil
import pkg_resources
from pkg_resources import Requirement, resource_filename, ResolutionError
import sys
import traceback
import subprocess
from io import open  # pylint: disable=redefined-builtin
from hashlib import md5

from khmer import reverse_complement as revcomp

import pytest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def _equals_rc(query, match):
    return (query == match) or (revcomp(query) == match)


def _contains_rc(match, query):
    return (query in match) or (revcomp(query) in match)


def _calc_md5(fp):
    m = md5()
    m.update(fp.read())
    return m.hexdigest()


def get_test_data(filename):
    filepath = None
    try:
        filepath = resource_filename(
            Requirement.parse("khmer"), "khmer/tests/test-data/" + filename)
    except ResolutionError:
        pass
    if not filepath or not os.path.isfile(filepath):
        filepath = os.path.join(os.path.dirname(__file__), 'test-data',
                                filename)
    return filepath


CLEANUPLIST = []


def get_temp_filename(filename, tempdir=None):
    if tempdir is None:
        tempdir = tempfile.mkdtemp(prefix='khmertest_')
    CLEANUPLIST.append(tempdir)

    return os.path.join(tempdir, filename)


def cleanup():
    global CLEANUPLIST  # pylint: disable=global-statement

    for path in CLEANUPLIST:
        shutil.rmtree(path, ignore_errors=True)
    CLEANUPLIST = []


def scriptpath(scriptname='interleave-reads.py'):
    """Return the path to the scripts, in both dev and install situations."""
    # note - it doesn't matter what the scriptname is here, as long as
    # it's some khmer script present in this version of khmer.

    path = os.path.join(os.path.dirname(__file__), "../scripts")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    path = os.path.join(os.path.dirname(__file__), "../../EGG-INFO/scripts")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    for path in os.environ['PATH'].split(':'):
        if os.path.exists(os.path.join(path, scriptname)):
            return path


def _runscript(scriptname, sandbox=False):
    """Find & run a script with exec (i.e. not via os.system or subprocess)."""
    namespace = {"__name__": "__main__"}
    namespace['sys'] = globals()['sys']

    try:
        pkg_resources.get_distribution("khmer").run_script(
            scriptname, namespace)
        return 0
    except pkg_resources.ResolutionError:
        pass

    if sandbox:
        path = os.path.join(os.path.dirname(__file__), "../sandbox")
    else:
        path = scriptpath()

    scriptfile = os.path.join(path, scriptname)
    if os.path.isfile(scriptfile):
        if os.path.isfile(scriptfile):
            exec(compile(open(scriptfile).read(), scriptfile, 'exec'),
                 namespace)
            return 0
    else:
        raise RuntimeError("Tried to execute {} but it is"
                           " not a file.".format(scriptfile))

    return -1


def runscript(scriptname, args, in_directory=None,
              fail_ok=False, sandbox=False):
    """Run a Python script using exec().

    Run the given Python script, with the given args, in the given directory,
    using 'exec'.  Mimic proper shell functionality with argv, and capture
    stdout and stderr.

    When using :attr:`fail_ok`=False in tests, specify the expected error.
    """
    sysargs = [scriptname]
    sysargs.extend(args)

    cwd = os.getcwd()

    try:
        status = -1
        oldargs = sys.argv
        sys.argv = sysargs

        oldout, olderr = sys.stdout, sys.stderr
        sys.stdout = StringIO()
        sys.stdout.name = "StringIO"
        sys.stderr = StringIO()

        if in_directory:
            os.chdir(in_directory)
        else:
            in_directory = cwd

        try:
            print('running:', scriptname, 'in:', in_directory, file=oldout)
            print('arguments', sysargs, file=oldout)

            status = _runscript(scriptname, sandbox=sandbox)
        except SystemExit as err:
            status = err.code
        except:  # pylint: disable=bare-except
            traceback.print_exc(file=sys.stderr)
            status = -1
    finally:
        sys.argv = oldargs
        out, err = sys.stdout.getvalue(), sys.stderr.getvalue()
        sys.stdout, sys.stderr = oldout, olderr

        os.chdir(cwd)

    if status != 0 and not fail_ok:
        print(out)
        print(err)
        assert False, (status, out, err)

    return status, out, err


def run_shell_cmd(cmd, fail_ok=False, in_directory=None):
    cwd = os.getcwd()
    if in_directory:
        os.chdir(in_directory)

    print('running: ', cmd)
    try:
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        (out, err) = proc.communicate()

        out = out.decode('utf-8')
        err = err.decode('utf-8')

        if proc.returncode != 0 and not fail_ok:
            print('out:', out)
            print('err:', err)
            raise AssertionError("exit code is non zero: %d" % proc.returncode)

        return (proc.returncode, out, err)
    finally:
        os.chdir(cwd)


def longify(listofints):
    """List of ints => list of longs, only on py2.

    Takes a list of numeric types, and returns longs on python2, or the
    original list on python3.
    """
    # For map(long, [list of ints]) cross-version hackery
    if sys.version_info.major < 3:
        return map(long, listofints)  # pylint: disable=bad-builtin
    return listofints


def copy_test_data(testfile, newfilename=None):
    basename = os.path.basename(testfile)
    if newfilename is not None:
        basename = newfilename
    infile = get_temp_filename(basename)
    shutil.copyfile(get_test_data(testfile), infile)
    return infile



# -*- coding: UTF-8 -*-
# cython: c_string_type=unicode, c_string_encoding=utf8

from __future__ import unicode_literals

from libc.stdint cimport uintptr_t 

from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.utility cimport pair
from libcpp.string cimport string

#from .wrapper cimport *

'''
extern declarations for liboxli.
'''

# C++ ostream wrapper code stolen shamelessly from stackoverflow
# http://stackoverflow.com/questions/30984078/cython-working-with-c-streams
# We need ostream to wrap ReadParser
cdef extern from "<iostream>" namespace "std":
    cdef cppclass ostream:
        ostream& write(const char*, int) except +

# obviously std::ios_base isn't a namespace, but this lets
# Cython generate the connect C++ code
cdef extern from "<iostream>" namespace "std::ios_base":
    cdef cppclass open_mode:
        pass
    cdef open_mode binary
    # you can define other constants as needed


cdef extern from "<fstream>" namespace "std":
    cdef cppclass ofstream(ostream):
        # constructors
        ofstream(const char*) except +
        ofstream(const char*, open_mode) except+


cdef extern from  "oxli/read_parsers.hh" namespace "oxli::read_parsers":
    cdef cppclass CpSequence "oxli::read_parsers::Read":
        string name
        string description
        string sequence
        string quality
        string cleaned_seq

        void reset()
        void write_fastx(ostream&)
        void set_cleaned_seq()        

    ctypedef pair[CpSequence,CpSequence] CpSequencePair \
        "oxli::read_parsers::ReadPair"

    cdef cppclass CpReadParser "oxli::read_parsers::ReadParser" [SeqIO]:
        CpReadParser(unique_ptr[SeqIO])
        CpReadParser(CpReadParser&)

        CpSequence get_next_read()
        CpSequencePair get_next_read_pair()
        CpSequencePair get_next_read_pair(uint8_t)

        uintptr_t get_num_reads()
        bool is_complete()
        void close()

    cdef cppclass CpFastxReader "oxli::read_parsers::FastxReader":
        CpFastxReader()
        CpFastxReader(const string&)
        CpFastxReader(CpFastxReader&)

        CpSequence get_next_read()
        bool is_complete()
        uintptr_t get_num_reads()
        void close()

    unique_ptr[CpReadParser[SeqIO]] get_parser[SeqIO](const string&) 
    ctypedef unique_ptr[CpReadParser[CpFastxReader]] FastxParserPtr


cdef extern from "oxli/alphabets.hh" namespace "oxli":
    cdef string DNA_SIMPLE "oxli::alphabets::DNA_SIMPLE"
    cdef string DNAN_SIMPLE "oxli::alphabets::DNAN_SIMPLE"
    cdef string RNA_SIMPLE "oxli::alphabets::RNA_SIMPLE"
    cdef string RNAN_SIMPLE "oxli::alphabets::RNAN_SIMPLE"
    cdef string IUPAC_NUCL "oxli::alphabets::IUPAC_NUCL"
    cdef string IUPAC_AA "oxli::alphabets::IUPAC_AA"

'''
Extension Classes wrapping liboxli.
'''

cdef class Alphabets:

    @staticmethod
    cdef string _get(string name)


cdef class Sequence:
    cdef CpSequence _obj

    @staticmethod
    cdef Sequence _wrap(CpSequence cseq)


cdef class ReadBundle:
    cdef list reads


cdef class FastxParser:
    cdef unique_ptr[CpReadParser[CpFastxReader]] _this

    cpdef bool is_complete(self)
    cdef Sequence _next(self)


cdef class SanitizedFastxParser(FastxParser):
    cdef readonly int n_bad
    cdef readonly string _alphabet
    cdef bool convert_n

    cpdef bool is_complete(self)
    cdef Sequence _next(self)


cdef class SplitPairedReader:

    cdef FastxParser left_parser
    cdef FastxParser right_parser
    cdef readonly int min_length
    cdef readonly bool force_name_match

    cdef tuple _next(self)


cdef class BrokenPairedReader:

    cdef FastxParser parser
    cdef readonly int min_length
    cdef readonly bool force_single
    cdef readonly bool require_paired
    cdef readonly Sequence record

    cdef tuple _next(self)


cpdef tuple _split_left_right(unicode s)

cdef tuple _cppstring_split_left_right(string& s)

cdef int _check_is_pair(Sequence first, Sequence second)

cpdef bool check_is_left(s)

cpdef bool check_is_right(s)

cdef inline bool is_valid(const char base, string& alphabet)

cdef inline bool sanitize_sequence(string& sequence,
                                   string& alphabet,
                                   bool convert_n)




from libcpp cimport bool
from libcpp.memory cimport shared_ptr
from libcpp.queue cimport queue
from libcpp.set cimport set
from libcpp.string cimport string

from oxli_types cimport *

cdef extern from "oxli/kmer_hash.hh" namespace "oxli":
    cdef cppclass CpKmer "oxli::Kmer":
        HashIntoType kmer_f
        HashIntoType kmer_r
        HashIntoType kmer_u

        CpKmer(HashIntoType, HashIntoType, HashIntoType)
        CpKmer(string, WordLength)
        CpKmer(const CpKmer&)
        CpKmer()

        bool is_forward() const
        void set_from_unique_hash(HashIntoType, WordLength)

    cdef cppclass CpKmerFactory "oxli::KmerFactory":
        KmerFactory(WordLength)

        CpKmer build_kmer(HashIntoType) const
        CpKmer build_kmer(HashIntoType, HashIntoType) const
        CpKmer build_kmer(string &) const
        CpKmer build_kmer(const char *) const

    cdef cppclass CpKmerIterator "oxli::KmerIterator" (CpKmerFactory):
        CpKmerIterator(const char *, unsigned char)
        CpKmer first(HashIntoType &, HashIntoType &)
        CpKmer next(HashIntoType &, HashIntoType &)
        CpKmer first()
        CpKmer next()
        bool done()
        unsigned int get_start_pos() const
        unsigned int get_end_pos() const

    HashIntoType _hash(const string, const WordLength)
    HashIntoType _hash(const string, const WordLength, 
                       HashIntoType &, HashIntoType &)
    HashIntoType _hash(const char *, const WordLength)
    HashIntoType _hash(const char *, const WordLength,
                       HashIntoType &, HashIntoType &)
    HashIntoType _hash_forward(const char *, WordLength)
    string _revhash(HashIntoType, WordLength)
    string _revcomp(const string&)
    HashIntoType _hash_murmur(const string&, const WordLength)
    HashIntoType _hash_murmur(const string&,
                              HashIntoType&, HashIntoType&)
    HashIntoType _hash_murmur_forward(const string&)


cdef extern from "oxli/oxli.hh" namespace "oxli":
    ctypedef queue[CpKmer] KmerQueue
    ctypedef set[CpKmer] KmerSet
    ctypedef bool (*KmerFilter) (CpKmer kmer)


cdef class Kmer:
    cdef shared_ptr[CpKmer] _this
    cdef readonly str kmer

    @staticmethod
    cdef Kmer wrap(CpKmer * cpkmer, WordLength K)



#! /usr/bin/env python
# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2011-2015, Michigan State University.
# Copyright (C) 2015, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=invalid-name,missing-docstring
"""
Interleave left and right reads.

Take two files containing left & right reads from a paired-end sequencing run,
and interleave them.

% scripts/interleave-reads.py <R1> <R2> [ -o <outputfile> ]

By default, output is sent to stdout; or use -o. Use '-h' for parameter help.
"""
from __future__ import print_function

import screed
import sys
import textwrap
from khmer import __version__
from khmer.kfile import check_input_files, check_space
from khmer.khmer_args import sanitize_help, KhmerArgumentParser
from khmer.khmer_args import FileType as khFileType
from khmer.kfile import (add_output_compression_type, get_file_writer,
                         describe_file_handle)
from khmer.utils import (write_record_pair, check_is_left, check_is_right,
                         check_is_pair)

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


def get_parser():
    epilog = """\
    The output is an interleaved set of reads, with each read in <R1> paired
    with a read in <R2>. By default, the output goes to stdout unless
    :option:`-o`/:option:`--output` is specified.

    As a "bonus", this file ensures that if read names are not already
    formatted properly, they are reformatted consistently, such that
    they look like the pre-1.8 Casava format (`@name/1`, `@name/2`).
    This reformatting can be switched off with the
    :option:`--no-reformat` flag.

    Example::

        interleave-reads.py tests/test-data/paired.fq.1 \\
                tests/test-data/paired.fq.2 -o paired.fq"""
    parser = KhmerArgumentParser(
        description='Produce interleaved files from R1/R2 paired files',
        epilog=textwrap.dedent(epilog))

    parser.add_argument('left')
    parser.add_argument('right')
    parser.add_argument('-o', '--output', metavar="filename",
                        type=khFileType('wb'),
                        default=sys.stdout)
    parser.add_argument('--no-reformat', default=False, action='store_true',
                        help='Do not reformat read names or enforce\
                              consistency')
    parser.add_argument('-f', '--force', default=False, action='store_true',
                        help='Overwrite output file if it exists')
    add_output_compression_type(parser)
    return parser


def main():
    args = sanitize_help(get_parser()).parse_args()

    check_input_files(args.left, args.force)
    check_input_files(args.right, args.force)
    check_space([args.left, args.right], args.force)

    s1_file = args.left
    s2_file = args.right

    print("Interleaving:\n\t%s\n\t%s" % (s1_file, s2_file), file=sys.stderr)

    outfp = get_file_writer(args.output, args.gzip, args.bzip)

    counter = 0
    screed_iter_1 = screed.open(s1_file)
    screed_iter_2 = screed.open(s2_file)
    for read1, read2 in zip_longest(screed_iter_1, screed_iter_2):
        if read1 is None or read2 is None:
            print(("ERROR: Input files contain different number"
                   " of records."), file=sys.stderr)
            sys.exit(1)

        if counter % 100000 == 0:
            print('...', counter, 'pairs', file=sys.stderr)
        counter += 1

        name1 = read1.name
        name2 = read2.name

        if not args.no_reformat:
            if not check_is_left(name1):
                name1 += '/1'
            if not check_is_right(name2):
                name2 += '/2'

            read1.name = name1
            read2.name = name2

            if not check_is_pair(read1, read2):
                print("ERROR: This doesn't look like paired data! "
                      "%s %s" % (read1.name, read2.name), file=sys.stderr)
                sys.exit(1)

        write_record_pair(read1, read2, outfp)

    print('final: interleaved %d pairs' % counter, file=sys.stderr)
    print('output written to', describe_file_handle(outfp), file=sys.stderr)


if __name__ == '__main__':
    main()



# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2014-2015, Michigan State University.
# Copyright (C) 2015-2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=missing-docstring,protected-access,no-member,invalid-name
from __future__ import division, print_function, unicode_literals
from __future__ import absolute_import

import khmer

from screed.fasta import fasta_iter

from . import khmer_tst_utils as utils
import pytest


K = 20  # size of kmer
ERR_RATE = 0.01
N_UNIQUE = 3960
TRANSLATE = {'A': 'T', 'C': 'G', 'T': 'A', 'G': 'C'}


def teardown():
    utils.cleanup()


def test_hll_add_python():
    # test python code to count unique kmers using HyperLogLog.
    # use the lower level add() method, which accepts anything,
    # and compare to an exact count using collections.Counter

    filename = utils.get_test_data('random-20-a.fa')
    hllcpp = khmer.HLLCounter(ERR_RATE, K)
    counter = set()

    for n, record in enumerate(fasta_iter(open(filename))):
        sequence = record['sequence']
        seq_len = len(sequence)
        for n in range(0, seq_len + 1 - K):
            kmer = sequence[n:n + K]
            rc = "".join(TRANSLATE[c] for c in kmer[::-1])

            hllcpp.add(kmer)

            if rc in counter:
                kmer = rc
            counter.update([kmer])

    n_unique = len(counter)

    assert n_unique == N_UNIQUE
    assert abs(1 - float(hllcpp.estimate_cardinality()) / N_UNIQUE) < ERR_RATE


def test_hll_consume_string():
    # test c++ code to count unique kmers using HyperLogLog,
    # using screed to feed each read to the counter.

    filename = utils.get_test_data('random-20-a.fa')
    hllcpp = khmer.HLLCounter(ERR_RATE, K)
    n_consumed = n = 0
    for n, record in enumerate(fasta_iter(open(filename)), 1):
        n_consumed += hllcpp.consume_string(record['sequence'])

    assert n == 99
    assert n_consumed == 3960
    assert abs(1 - float(hllcpp.estimate_cardinality()) / N_UNIQUE) < ERR_RATE


def test_hll_empty_fasta():
    filename = utils.get_test_data('test-empty.fa')
    hll = khmer.HLLCounter(ERR_RATE, K)
    with pytest.raises(OSError):
        hll.consume_seqfile(filename)


def test_hll_consume_seqfile():
    # test c++ code to count unique kmers using HyperLogLog

    filename = utils.get_test_data('random-20-a.fa')
    hllcpp = khmer.HLLCounter(ERR_RATE, K)
    n, n_consumed = hllcpp.consume_seqfile(filename)

    assert n == 99
    assert n_consumed == 3960
    assert abs(1 - float(hllcpp.estimate_cardinality()) / N_UNIQUE) < ERR_RATE


def test_hll_consume_seqfile_ep():
    # During estimation trigger the _Ep() method,
    # we need all internal counters values to be different than zero for this.

    filename = utils.get_test_data('paired-mixed.fa')
    hll = khmer.HLLCounter(0.36, 32)
    n, n_consumed = hll.consume_seqfile(filename)

    assert all(c != 0 for c in hll.counters)
    assert len(hll) == 236
    assert n == 11
    assert n_consumed == 575


def test_hll_consume_seqfile_estimate_bias():
    # During estimation trigger the estimate_bias method,
    # we need all internal counters values to be different than zero for this,
    # and also the cardinality should be small (if it is large we fall on the
    # default case).

    filename = utils.get_test_data("test-abund-read-3.fa")
    hll = khmer.HLLCounter(0.36, K)
    n, n_consumed = hll.consume_seqfile(filename)

    assert all(c != 0 for c in hll.counters)
    assert len(hll) == 79
    assert n == 21
    assert n_consumed == 1176


def test_hll_len():
    filename = utils.get_test_data('random-20-a.fa')
    hllcpp = khmer.HLLCounter(ERR_RATE, K)
    n, n_consumed = hllcpp.consume_seqfile(filename)

    assert n == 99
    assert n_consumed == 3960
    assert hllcpp.estimate_cardinality() == len(hllcpp)


def test_hll_empty():
    hllcpp = khmer.HLLCounter(ERR_RATE, K)

    assert len(hllcpp) == 0


def test_hll_readonly_alpha():
    hllcpp = khmer.HLLCounter(ERR_RATE, K)
    with pytest.raises(AttributeError):
        hllcpp.alpha = 5


def test_hll_cover_calc_alpha():
    hllcpp = khmer.HLLCounter(0.36, K)
    counters = hllcpp.counters
    assert hllcpp.alpha == 0.673
    assert len(counters) == 2 ** 4

    hllcpp = khmer.HLLCounter(0.21, K)
    counters = hllcpp.counters
    assert hllcpp.alpha == 0.697
    assert len(counters) == 2 ** 5

    hllcpp = khmer.HLLCounter(0.16, K)
    counters = hllcpp.counters
    assert hllcpp.alpha == 0.709
    assert len(counters) == 2 ** 6


def test_hll_invalid_base():
    hllcpp = khmer.HLLCounter(ERR_RATE, 5)

    # this should succeed; invalid bases need to be removed before
    # hashing.
    hllcpp.consume_string("ACGTTTCGNAATNNNNN")


def test_hll_invalid_error_rate():
    # test if error_rate is a valid value

    with pytest.raises(ValueError):
        khmer.HLLCounter(-0.01, K)


def test_hll_invalid_error_rate_max():
    # test if error_rate is a valid value

    with pytest.raises(ValueError):
        khmer.HLLCounter(0.367696, K)


def test_hll_error_rate_max():
    # test if error_rate is a valid value

    hllcpp = khmer.HLLCounter(0.367695, K)
    assert len(hllcpp.counters) == 2 ** 4


def test_hll_invalid_error_rate_min():
    # test if error_rate is a valid value

    with pytest.raises(ValueError):
        khmer.HLLCounter(0.0040624, K)


def test_hll_error_rate_min():
    # test if error_rate is a valid value

    hllcpp = khmer.HLLCounter(0.0040625, K)
    assert len(hllcpp.counters) == 2 ** 16


def test_hll_change_error_rate():
    hllcpp = khmer.HLLCounter(0.0040625, K)
    assert hllcpp.error_rate == 0.0040625

    # error rate is discrete, what we test here is if an error rate of 1%
    # rounds to the appropriate value
    hllcpp.error_rate = 0.01
    assert hllcpp.error_rate == 0.008125

    with pytest.raises(TypeError):
        del hllcpp.error_rate

    with pytest.raises(TypeError):
        hllcpp.error_rate = 5

    with pytest.raises(ValueError):
        hllcpp.error_rate = 2.5

    with pytest.raises(ValueError):
        hllcpp.error_rate = -10.

    # error rate can only be changed prior to first counting,
    hllcpp.consume_string('AAACCACTTGTGCATGTCAGTGCAGTCAGT')
    with pytest.raises(AttributeError):
        hllcpp.error_rate = 0.3


def test_hll_change_ksize():
    hllcpp = khmer.HLLCounter(0.0040625, K)
    assert hllcpp.ksize == K

    hllcpp.ksize = 24
    assert hllcpp.ksize == 24

    hllcpp.ksize = 12
    assert hllcpp.ksize == 12

    with pytest.raises(ValueError):
        hllcpp.ksize = -20

    with pytest.raises(TypeError):
        del hllcpp.ksize

    with pytest.raises(TypeError):
        hllcpp.ksize = 33.4

    # error rate can only be changed prior to first counting,
    hllcpp.consume_string('AAACCACTTGTGCATGTCAGTGCAGTCAGT')
    with pytest.raises(AttributeError):
        hllcpp.ksize = 30


def test_hll_get_counters():
    hll = khmer.HLLCounter(0.36, K)
    counters = hll.counters
    assert len(counters) == 2 ** 4
    assert all(c == 0 for c in counters)


def test_hll_merge_1():
    hll = khmer.HLLCounter(0.36, K)
    hll2 = khmer.HLLCounter(0.36, K - 1)

    try:
        hll.merge(hll2)
        assert 0, "previous statement should fail with a ValueError"
    except ValueError as err:
        print(str(err))


def test_hll_merge_2():
    hll = khmer.HLLCounter(0.10, K)
    hll2 = khmer.HLLCounter(0.36, K)

    try:
        hll.merge(hll2)
        assert 0, "previous statement should fail with a ValueError"
    except ValueError as err:
        print(str(err))


def test_hll_merge_3():
    hll = khmer.HLLCounter(0.36, 32)
    hll2 = khmer.HLLCounter(0.36, 32)

    filename = utils.get_test_data('paired-mixed.fa')
    hll = khmer.HLLCounter(0.36, 32)
    hll.consume_seqfile(filename)

    hll2 = khmer.HLLCounter(0.36, 32)
    hll2.consume_seqfile(filename)

    assert len(hll) == 236
    assert len(hll2) == 236

    hll.merge(hll2)
    assert len(hll) == 236



# -*- coding: UTF-8 -*-
# cython: c_string_type=unicode, c_string_encoding=utf8

from libcpp.string cimport string
from libcpp.memory cimport make_shared
from libc.stdint cimport uint64_t
from cython.operator cimport dereference as deref

from oxli_types cimport *

cdef class Kmer:

    def __cinit__(self, str kmer=None):
        self.kmer = kmer
        if self.kmer is not None:
            self._this.reset(new CpKmer(kmer.encode('utf-8'), len(kmer)))
        else:
            self._this.reset(new CpKmer())

    def __len__(self):
        return len(self.kmer)

    def __str__(self):
        return self.kmer

    def __hash__(self):
        return self.kmer_u

    def __repr__(self):
        return self.kmer

    @property
    def kmer_f(self):
        return deref(self._this).kmer_f

    @property
    def kmer_r(self):
        return deref(self._this).kmer_r

    @property
    def kmer_u(self):
        return deref(self._this).kmer_u

    def reverse_complement(self):
        cdef HashIntoType tmp = deref(self._this).kmer_f
        deref(self._this).kmer_f = deref(self._this).kmer_r
        deref(self._this).kmer_r = tmp
        self.kmer = _revcomp(self.kmer.encode('utf-8'))

    @property
    def is_forward(self):
        return deref(self._this).is_forward()

    @staticmethod
    cdef Kmer wrap(CpKmer * cpkmer, WordLength K):
        cdef Kmer kmer = Kmer()
        kmer._this.reset(cpkmer)
        kmer.kmer = _revhash(kmer.kmer_u, K)
        return kmer

    @staticmethod
    def create(HashIntoType tag, WordLength K):
        cdef Kmer kmer = Kmer()
        deref(kmer._this).set_from_unique_hash(tag, K)
        kmer.kmer = _revhash(kmer.kmer_u, K)
        return kmer



# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2010-2015, Michigan State University.
# Copyright (C) 2015-2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=C0111,C0103
"""
Test code for HashSet objects.
"""
from __future__ import print_function
from __future__ import absolute_import

import khmer
from . import khmer_tst_utils as utils


def test_bad_construct():
    try:
        hs = khmer.HashSet()
        assert 0, "HashSet constructor should fail w/o argument"
    except TypeError:
        pass

    try:
        hs = khmer.HashSet(5, [{}])
        assert 0, "HashSet constructor should fail w/o list of k-mers"
    except ValueError:
        pass


def test_iter_single():
    hs = khmer.HashSet(5, [6])
    for k in hs:
        assert k == 6
        print(k)


def test_iter_double():
    x = [6, 9, 20]
    hs = khmer.HashSet(5, x)
    for i, k in enumerate(hs):
        assert k == x[i], (k, x[i])


def test_iter_single():
    hs = khmer.HashSet(5, [6])
    k = iter(hs)
    k2 = iter(k)
    assert k == k2


def test_add():
    hs = khmer.HashSet(5)
    hs.add(7)
    hs.add(4)

    assert list(sorted(hs)) == [4, 7]


def test_update():
    hs = khmer.HashSet(5)
    x = [5, 10, 15, 2**35]
    hs.update(x)

    assert list(sorted(hs)) == [5, 10, 15, 2**35]


def test_update_bad():
    hs = khmer.HashSet(5)
    x = [5, 10, 15, 2**35, {}]
    try:
        hs.update(x)
        assert 0, "cannot add dict to a HashSet"
    except ValueError:
        pass


def test_remove():
    hs = khmer.HashSet(5, [8, 10])
    assert len(hs) == 2
    hs.remove(8)
    assert len(hs) == 1
    assert list(hs) == [10]


def test_remove_2():
    hs = khmer.HashSet(5, [8, 10])
    assert len(hs) == 2
    try:
        hs.remove(15)
        assert 0, "hs.remove should raise an Exception"
    except ValueError:
        pass
    assert len(hs) == 2
    assert list(sorted(hs)) == [8, 10]


def test_contains_1():
    hs = khmer.HashSet(5, [8, 10])
    assert 8 in hs
    assert 10 in hs
    assert 2**35 not in hs


def test_concat_1():
    hs = khmer.HashSet(5, [10, 12])
    hs2 = khmer.HashSet(5, [10, 13])

    hs3 = hs + hs2
    assert list(sorted(hs3)) == [10, 12, 13]


def test_concat_2():
    hs = khmer.HashSet(5, [10, 12])
    hs2 = khmer.HashSet(5, [10, 13])

    hs += hs2
    assert list(sorted(hs)) == [10, 12, 13]


def test_concat_1_fail():
    hs = khmer.HashSet(5, [10, 12])
    hs2 = khmer.HashSet(4, [10, 13])

    try:
        hs3 = hs + hs2
        assert 0, "concat should fail - different ksize"
    except ValueError:
        pass


def test_concat_2_fail():
    hs = khmer.HashSet(5, [10, 12])
    hs2 = khmer.HashSet(4, [10, 13])

    try:
        hs += hs2
        assert 0, "inplace concat should fail - different ksize"
    except ValueError:
        pass



# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2010-2015, Michigan State University.
# Copyright (C) 2015-2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=too-few-public-methods,no-init,missing-docstring
"""This is khmer; please see http://khmer.readthedocs.io/."""


from __future__ import print_function
from collections import namedtuple
from math import log
import json

from khmer._khmer import Countgraph as _Countgraph
from khmer._khmer import SmallCountgraph as _SmallCountgraph
from khmer._khmer import Counttable as _Counttable
from khmer._khmer import SmallCounttable as _SmallCounttable
from khmer._khmer import GraphLabels as _GraphLabels
from khmer._khmer import Nodegraph as _Nodegraph
from khmer._khmer import Nodetable as _Nodetable
from khmer._khmer import HLLCounter as _HLLCounter
from khmer._khmer import ReadAligner as _ReadAligner

from khmer._khmer import HashSet
from khmer._khmer import Read
from khmer._khmer import forward_hash
# tests/test_{functions,countgraph,counting_single}.py

from khmer._khmer import forward_hash_no_rc  # tests/test_functions.py

from khmer._khmer import reverse_hash  # tests/test_functions.py
# tests/counting_single.py

from khmer._khmer import hash_murmur3        # tests/test_functions.py
from khmer._khmer import hash_no_rc_murmur3  # tests/test_functions.py

from khmer._khmer import reverse_complement

from khmer._khmer import get_version_cpp as __version_cpp__
# tests/test_version.py

from khmer._khmer import ReadParser  # sandbox/to-casava-1.8-fastq.py
# tests/test_read_parsers.py,scripts/{filter-abund-single,load-graph}.py
# scripts/{abundance-dist-single,load-into-counting}.py

from khmer._khmer import FILETYPES

import sys

from struct import pack, unpack

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


_buckets_per_byte = {
    'countgraph': 1,
    'smallcountgraph': 2,
    'nodegraph': 8,
}


def load_nodegraph(filename):
    """Load a nodegraph object from the given filename and return it.

    Keyword argument:
    filename -- the name of the nodegraph file
    """
    nodegraph = _Nodegraph(1, [1])
    nodegraph.load(filename)

    return nodegraph


def load_nodetable(filename):
    """Load a nodetable object from the given filename and return it.

    Keyword argument:
    filename -- the name of the nodegraph file
    """
    nodetable = _Nodetable(1, [1])
    nodetable.load(filename)

    return nodetable


def load_countgraph(filename, small=False):
    """Load a countgraph object from the given filename and return it.

    Keyword argument:
    filename -- the name of the countgraph file
    small -- set this to load a SmallCountgraph instance
    """
    if small:
        countgraph = _SmallCountgraph(1, [1])
        countgraph.load(filename)

    else:
        countgraph = _Countgraph(1, [1])
        countgraph.load(filename)

    return countgraph


def load_counttable(filename, small=False):
    """Load a counttable object from the given filename and return it.

    Keyword argument:
    filename -- the name of the counttable file
    small -- set this to load a SmallCounttable instance
    """
    if small:
        counttable = _SmallCounttable(1, [1])
        counttable.load(filename)

    else:
        counttable = _Counttable(1, [1])
        counttable.load(filename)

    return counttable


def extract_nodegraph_info(filename):
    """Open the given nodegraph file and return a tuple of information.

    Returns: the k-mer size, the table size, the number of tables, the version
    of the table format, and the type of table flag.

    Keyword argument:
    filename -- the name of the nodegraph file to inspect
    """
    ksize = None
    n_tables = None
    table_size = None
    signature = None
    version = None
    ht_type = None
    occupied = None

    uint_size = len(pack('I', 0))
    uchar_size = len(pack('B', 0))
    ulonglong_size = len(pack('Q', 0))

    try:
        with open(filename, 'rb') as nodegraph:
            signature, = unpack('4s', nodegraph.read(4))
            version, = unpack('B', nodegraph.read(1))
            ht_type, = unpack('B', nodegraph.read(1))
            ksize, = unpack('I', nodegraph.read(uint_size))
            n_tables, = unpack('B', nodegraph.read(uchar_size))
            occupied, = unpack('Q', nodegraph.read(ulonglong_size))
            table_size, = unpack('Q', nodegraph.read(ulonglong_size))
        if signature != b"OXLI":
            raise ValueError("Node graph '{}' is missing file type "
                             "signature".format(filename) + str(signature))
    except:
        raise ValueError("Node graph '{}' is corrupt ".format(filename))

    return ksize, round(table_size, -2), n_tables, version, ht_type, occupied


def extract_countgraph_info(filename):
    """Open the given countgraph file and return a tuple of information.

    Return: the k-mer size, the table size, the number of tables, the bigcount
    flag, the version of the table format, and the type of table flag.

    Keyword argument:
    filename -- the name of the countgraph file to inspect
    """
    CgInfo = namedtuple("CgInfo", ['ksize', 'n_tables', 'table_size',
                                   'use_bigcount', 'version', 'ht_type',
                                   'n_occupied'])
    ksize = None
    n_tables = None
    table_size = None
    signature = None
    version = None
    ht_type = None
    use_bigcount = None
    occupied = None

    uint_size = len(pack('I', 0))
    ulonglong_size = len(pack('Q', 0))

    try:
        with open(filename, 'rb') as countgraph:
            signature, = unpack('4s', countgraph.read(4))
            version, = unpack('B', countgraph.read(1))
            ht_type, = unpack('B', countgraph.read(1))
            if ht_type != FILETYPES['SMALLCOUNT']:
                use_bigcount, = unpack('B', countgraph.read(1))
            else:
                use_bigcount = None
            ksize, = unpack('I', countgraph.read(uint_size))
            n_tables, = unpack('B', countgraph.read(1))
            occupied, = unpack('Q', countgraph.read(ulonglong_size))
            table_size, = unpack('Q', countgraph.read(ulonglong_size))
        if signature != b'OXLI':
            raise ValueError("Count graph file '{}' is missing file type "
                             "signature. ".format(filename) + str(signature))
    except:
        raise ValueError("Count graph file '{}' is corrupt ".format(filename))

    return CgInfo(ksize, n_tables, round(table_size, -2), use_bigcount,
                  version, ht_type, occupied)


def calc_expected_collisions(graph, force=False, max_false_pos=.2):
    """Do a quick & dirty expected collision rate calculation on a graph.

    Also check to see that collision rate is within threshold.

    Keyword argument:
    graph: the countgraph or nodegraph object to inspect
    """
    sizes = graph.hashsizes()
    n_ht = float(len(sizes))
    occupancy = float(graph.n_occupied())
    min_size = min(sizes)

    fp_one = occupancy / min_size
    fp_all = fp_one ** n_ht

    if fp_all > max_false_pos:
        print("**", file=sys.stderr)
        print("** ERROR: the graph structure is too small for ",
              file=sys.stderr)
        print("** this data set.  Increase data structure size",
              file=sys.stderr)
        print("** with --max_memory_usage/-M.", file=sys.stderr)
        print("**", file=sys.stderr)
        print("** Do not use these results!!", file=sys.stderr)
        print("**", file=sys.stderr)
        print("** (estimated false positive rate of %.3f;" % fp_all,
              file=sys.stderr, end=' ')
        print("max recommended %.3f)" % max_false_pos, file=sys.stderr)
        print("**", file=sys.stderr)

        if not force:
            sys.exit(1)

    return fp_all


def is_prime(number):
    """Check if a number is prime."""
    if number < 2:
        return False
    if number == 2:
        return True
    if number % 2 == 0:
        return False
    for _ in range(3, int(number ** 0.5) + 1, 2):
        if number % _ == 0:
            return False
    return True


def get_n_primes_near_x(number, target):
    """Backward-find primes smaller than target.

    Step backwards until a number of primes (other than 2) have been
    found that are smaller than the target and return them.

    Keyword arguments:
    number -- the number of primes to find
    target -- the number to step backwards from
    """
    if target == 1 and number == 1:
        return [1]

    primes = []
    i = target - 1
    if i % 2 == 0:
        i -= 1
    while len(primes) != number and i > 0:
        if is_prime(i):
            primes.append(int(i))
        i -= 2

    if len(primes) != number:
        raise RuntimeError("unable to find %d prime numbers < %d" % (number,
                                                                     target))

    return primes


# Expose the cpython objects with __new__ implementations.
# These constructors add the functionality provided by the existing
# factory methods to the constructors defined over in cpython land.
# Additional functionality can be added to these classes as appropriate.


class Countgraph(_Countgraph):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        countgraph = _Countgraph.__new__(cls, k, primes)
        countgraph.primes = primes
        return countgraph


class SmallCountgraph(_SmallCountgraph):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        countgraph = _SmallCountgraph.__new__(cls, k, primes)
        countgraph.primes = primes
        return countgraph


class Counttable(_Counttable):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        counttable = _Counttable.__new__(cls, k, primes)
        counttable.primes = primes
        return counttable


class SmallCounttable(_SmallCounttable):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        counttable = _SmallCounttable.__new__(cls, k, primes)
        counttable.primes = primes
        return counttable


class GraphLabels(_GraphLabels):

    def __new__(cls, k, starting_size, n_tables):
        nodegraph = Nodegraph(k, starting_size, n_tables)
        graphlabels = _GraphLabels.__new__(cls, nodegraph)
        graphlabels.graph = nodegraph
        return graphlabels


class CountingGraphLabels(_GraphLabels):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        countgraph = _Countgraph(k, primes)
        class_ = _GraphLabels.__new__(cls, countgraph)
        class_.graph = countgraph
        return class_


class Nodegraph(_Nodegraph):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        nodegraph = _Nodegraph.__new__(cls, k, primes)
        nodegraph.primes = primes
        return nodegraph


class Nodetable(_Nodetable):

    def __new__(cls, k, starting_size, n_tables):
        primes = get_n_primes_near_x(n_tables, starting_size)
        nodetable = _Nodetable.__new__(cls, k, primes)
        nodetable.primes = primes
        return nodetable


class HLLCounter(_HLLCounter):
    """HyperLogLog counter.

    A HyperLogLog counter is a probabilistic data structure specialized on
    cardinality estimation.
    There is a precision/memory consumption trade-off: error rate determines
    how much memory is consumed.

    # Creating a new HLLCounter:

    >>> khmer.HLLCounter(error_rate, ksize)

    where the default values are:
      - error_rate: 0.01
      - ksize: 20
    """

    def __len__(self):
        """Return the cardinality estimate."""
        return _HLLCounter.estimate_cardinality(self)


class ReadAligner(_ReadAligner):
    """Sequence to graph aligner.

    ReadAligner uses a Countgraph (the counts of k-mers in the target DNA
    sequences) as an implicit De Bruijn graph. Input DNA sequences are aligned
    to this graph via a paired Hidden Markov Model.

    The HMM is configured upon class instantiation; default paramaters for the
    HMM are provided in 'defaultTransitionProbablitites' and
    'defaultScoringMatrix'.

    The main method is 'align'.
    """

    defaultTransitionProbabilities = (  # _M, _Ir, _Ig, _Mu, _Iru, _Igu
        (log(0.9848843, 2), log(0.0000735, 2), log(0.0000334, 2),
         log(0.0150068, 2), log(0.0000017, 2), log(0.0000003, 2)),  # M_
        (log(0.5196194, 2), log(0.4647955, 2), log(0.0059060, 2),
         log(0.0096792, 2)),  # Ir_
        (log(0.7611255, 2), log(0.2294619, 2), log(0.0072673, 2),
         log(0.0021453, 2)),  # Ig_
        (log(0.0799009, 2), log(0.0000262, 2), log(0.0001836, 2),
         log(0.9161349, 2), log(0.0033370, 2), log(0.0004173, 2)),  # Mu_
        (log(0.1434529, 2), log(0.0036995, 2), log(0.2642928, 2),
         log(0.5885548, 2)),  # Iru_
        (log(0.1384551, 2), log(0.0431328, 2), log(0.6362921, 2),
         log(0.1821200, 2))  # Igu_
    )

    defaultScoringMatrix = [
        log(0.955, 2), log(0.04, 2), log(0.004, 2), log(0.001, 2)]

    def __new__(cls, count_graph, trusted_cov_cutoff, bits_theta,
                **kwargs):

        if 'filename' in kwargs:
            with open(kwargs.pop('filename')) as paramfile:
                params = json.load(paramfile)
            scoring_matrix = params['scoring_matrix']
            transition_probabilities = params['transition_probabilities']
        else:
            if 'scoring_matrix' in kwargs:
                scoring_matrix = kwargs.pop('scoring_matrix')
            else:
                scoring_matrix = ReadAligner.defaultScoringMatrix
            if 'transition_probabilities' in kwargs:
                transition_probabilities = kwargs.pop(
                    'transition_probabilities')
            else:
                transition_probabilities = \
                    ReadAligner.defaultTransitionProbabilities
        readaligner = _ReadAligner.__new__(
            cls, count_graph, trusted_cov_cutoff, bits_theta, scoring_matrix,
            transition_probabilities)
        readaligner.graph = count_graph
        return readaligner

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Initialize ReadAligner.

        HMM state notation abbreviations:
        M_t - trusted match; M_u - untrusted match
        Ir_t - trusted read insert; Ir_u - untrusted read insert
        Ig_t - trusted graph insert; Ig_u - untrusted graph insert

        Keyword arguments:
        filename - a path to a JSON encoded file providing the scoring matrix
            for the HMM in an entry named 'scoring_matrix' and the transition
            probabilities for the HMM in an entry named
            'transition_probabilities'. If provided the remaining keyword
            arguments are ignored. (default: None)
        scoring_matrix - a list of floats: trusted match, trusted mismatch,
            unstrusted match, untrusted mismatch. (default:
                ReadAligner.defaultScoringMatrix)
        transition_probabilities - A sparse matrix as a tuple of six tuples.
            The inner tuples contain 6, 4, 4, 6, 4, and 4 floats respectively.
            Transition are notated as 'StartState-NextState':
            (
              ( M_t-M_t,  M_t-Ir_t,  M_t-Ig_t,  M_t-M_u,  M_t-Ir_u,  M_t-Ig_u),
              (Ir_t-M_t, Ir_t-Ir_t,            Ir_t-M_u, Ir_t-Ir_u           ),
              (Ig_t-M_t,          , Ig_t-Ig_t, Ig_t-M_u,            Ig_t-Ig_u),
              ( M_u-M_t,  M_u-Ir_t,  M_u-Ig_t,  M_u-M_u,  M_u-Ir_u,  M_u-Ig_u),
              (Ir_u-M_t, Ir_u-Ir_t,            Ir_u-M_u, Ir_u-Ir_u           ),
              (Ig_u-M_t,          , Ig_u-Ig_t, Ig_u-M_u,            Ig_u-Ig_u)
            )
            (default: ReadAligner.defaultTransitionProbabilities)


        Note: the underlying CPython implementation creates the ReadAligner
        during the __new__ process and so the class initialization actually
        occurs there. Instatiation is documented here in __init__ as this is
        the traditional way.
        """
        _ReadAligner.__init__(self)

from khmer._oxli.assembly import (LinearAssembler, SimpleLabeledAssembler,
                                  JunctionCountAssembler)



# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org

from __future__ import print_function
from __future__ import absolute_import

import random

from khmer import SmallCounttable

from . import khmer_tst_utils as utils


# This mostly tests the underlying NibbleStorage class

def test_single_add():
    sct = SmallCounttable(4, 1e6, 4)
    sct.add("AAAA")
    assert sct.get("AAAA") == 1


def test_split_byte_murmur():
    # check the byte is correctly split when using murmur hash
    sct = SmallCounttable(4, 4, 1)

    # these kmers were carefully chosen to have hash values that produce
    # consecutive indices in the count table.
    a = "AAAC"
    b = "AAAG"

    assert sct.get_kmer_hashes(a) == [11898086063751343884]
    assert sct.get_kmer_hashes(b) == [10548630838975263317]

    sct.add(a)

    assert sct.get(a) == 1
    assert sct.get(b) == 0


def test_overflow():
    # check that we do not overflow into other part of the byte
    sct = SmallCounttable(4, 1e6, 4)
    a = "AAAA"
    b = "AAAT"

    # try to overflow our 4bit counter
    for n in range(17):
        sct.add(a)

    assert sct.get(a) == 15
    assert sct.get(b) == 0

    # repeat with the other kmer that hashes to the other half of the byte
    sct = SmallCounttable(4, 1e6, 4)
    a = "AAAA"
    b = "AAAT"

    # try to overflow our 4bit counter
    for n in range(17):
        sct.add(b)

    assert sct.get(b) == 15
    assert sct.get(a) == 0


def test_random_kmers():
    # check for out-of-bounds errors and similar with random kmers
    rng = random.Random(1)

    sct = SmallCounttable(20, 1e2, 4)

    kmers = ["".join(rng.choice("ACGT") for _ in range(20))
             for n in range(400)]
    for kmer in kmers:
        sct.add(kmer)

    for kmer in kmers:
        sct.get(kmer)


def test_read_write():
    rng = random.Random(1)

    sct = SmallCounttable(20, 1e2, 4)

    kmers = ["".join(rng.choice("ACGT") for _ in range(20))
             for n in range(400)]
    for kmer in kmers:
        sct.add(kmer)

    fname = utils.get_temp_filename('zzz')

    sct.save(fname)

    # on purpose choose parameters that are different from sct
    sct2 = SmallCounttable(3, 1e4, 2)
    sct2.load(fname)
    assert sct.ksize() == sct2.ksize()
    for kmer in kmers:
        assert sct.get(kmer) == sct2.get(kmer)



#! /usr/bin/env python
# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2011-2015, Michigan State University.
# Copyright (C) 2015, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=invalid-name,missing-docstring
"""
Sequence trimming using stoptags.

Trim sequences at k-mers in the given stoptags file.  Output sequences
will be placed in 'infile.stopfilt'.

% python scripts/filter-stoptags.py <stoptags> <data1> [ <data2> <...> ]

Use '-h' for parameter help.
"""
from __future__ import print_function

import os
import textwrap
import sys
from khmer import Nodegraph
from khmer.thread_utils import ThreadedSequenceProcessor, verbose_loader
from khmer.kfile import check_input_files, check_space
from khmer.khmer_args import sanitize_help, KhmerArgumentParser

# @CTB K should be loaded from file...
DEFAULT_K = 32


def get_parser():
    epilog = """\
    Load stoptags in from the given `.stoptags` file and use them to trim
    or remove the sequences in `<file1-N>`.  Trimmed sequences will be placed
    in `<fileN>.stopfilt`.
    """
    parser = KhmerArgumentParser(
        description="Trim sequences at stoptags.",
        epilog=textwrap.dedent(epilog), citations=['graph'])
    parser.add_argument('-k', '--ksize', default=DEFAULT_K, type=int,
                        help='k-mer size')
    parser.add_argument('stoptags_file', metavar='input_stoptags_filename')
    parser.add_argument('input_filenames', metavar='input_sequence_filename',
                        nargs='+')
    parser.add_argument('-f', '--force', default=False, action='store_true',
                        help='Overwrite output file if it exists')
    return parser


def main():
    args = sanitize_help(get_parser()).parse_args()
    stoptags = args.stoptags_file
    infiles = args.input_filenames

    for _ in infiles:
        check_input_files(_, args.force)

    check_space(infiles, args.force)

    print('loading stop tags, with K', args.ksize, file=sys.stderr)
    nodegraph = Nodegraph(args.ksize, 1, 1)
    nodegraph.load_stop_tags(stoptags)

    def process_fn(record):
        name = record.name
        seq = record.sequence
        if 'N' in seq:
            return None, None

        trim_seq, trim_at = nodegraph.trim_on_stoptags(seq)

        if trim_at >= args.ksize:
            return name, trim_seq

        return None, None

    # the filtering loop
    for infile in infiles:
        print('filtering', infile, file=sys.stderr)
        outfile = os.path.basename(infile) + '.stopfilt'

        outfp = open(outfile, 'w')

        tsp = ThreadedSequenceProcessor(process_fn)
        tsp.start(verbose_loader(infile), outfp)

        print('output in', outfile, file=sys.stderr)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2013-2015, Michigan State University.
# Copyright (C) 2015, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
"""
Calculate the mismatch error profile for shotgun data, using a subset of
reads.  The output is placed in <infile>.errhist in the cwd by default.

% sandbox/calc-error-profile.py [ -o outfile ] <infile>

Reads FASTQ and FASTA input.
"""
from __future__ import division
from __future__ import print_function

import sys
import argparse
import khmer
import screed
import os.path

N_HT = 4
HASHSIZE = 1e7
K = 20
C = 10

MAX_SEQ_LEN = 65535
MAX_READS = 1e8
CHECK_EXIT = 25000


def exit_condition(n_consumed, n_checked):
    return (n_checked >= n_consumed or
            n_checked > 2e5)


def main():
    parser = argparse.ArgumentParser(
        "Calculate read error profile based on k-mer "
        "abundances of shotgun data.")

    parser.add_argument('filenames', nargs='+')
    parser.add_argument('-o', '--output', dest='output_file',
                        help="output file for histogram; defaults to "
                             "<first filename>.errhist in cwd.",
                        type=argparse.FileType('w'), default=None)
    parser.add_argument('--errors-per-read', dest='errors_per_read',
                        type=argparse.FileType('w'), default=None)

    args = parser.parse_args()

    #
    # Figure out what the output filename is going to be
    #

    output_file = args.output_file
    if output_file:
        output_filename = output_file.name
    else:
        filename = args.filenames[0]
        output_filename = os.path.basename(filename) + '.errhist'
        output_file = open(output_filename, 'w')

    # Start!

    # build a small counting hash w/default parameters. In general there
    # should be no need to change these parameters.
    ht = khmer.Countgraph(K, HASHSIZE, N_HT)

    # initialize list to contain counts of errors by position
    positions = [0] * MAX_SEQ_LEN
    lengths = []                  # keep track of sequence lengths

    n_consumed = 0
    bp_consumed = 0
    total = 0
    n_checked = 0

    # run through all the files; pick out reads; once they saturate,
    # look for errors.
    total = 0
    for filename in args.filenames:
        print('opening', filename, file=sys.stderr)

        for n, record in enumerate(screed.open(filename)):
            total += 1

            if total % CHECK_EXIT == 0:
                print('...', total, n_consumed, n_checked, file=sys.stderr)

                # two exit conditions: first, have we hit our max reads limit?
                if total >= MAX_READS:
                    break

                # OR, alternatively, have we counted enough reads?
                if exit_condition(n_consumed, n_checked):
                    break

            # for each sequence, calculate its coverage:
            seq = record.sequence.replace('N', 'A')
            med, _, _ = ht.get_median_count(seq)

            # if the coverage is unsaturated, consume.
            if med < C:
                ht.consume(seq)
                n_consumed += 1
                bp_consumed += len(seq)
            else:
                # for saturated data, find low-abund k-mers
                posns = ht.find_spectral_error_positions(seq, 2)
                lengths.append(len(seq))

                if args.errors_per_read:
                    print(record.name, \
                        ",".join(map(str, posns)), file=args.errors_per_read)

                # track the positions => errors
                for p in posns:
                    positions[p] += 1

                n_checked += 1

    # normalize for length
    lengths.sort()
    max_length = lengths[-1]

    length_count = [0] * max_length
    for j in range(max_length):
        length_count[j] = sum([1 for i in lengths if i >= j])

    # write!
    output_file.write('position error_count error_fraction\n')
    for n, i in enumerate(positions[:max_length]):
        print(n, i, float(i) / float(length_count[n]), file=output_file)

    output_file.close()

    print('', file=sys.stderr)
    print('total sequences:', total, file=sys.stderr)
    print('n consumed:', n_consumed, file=sys.stderr)
    print('n checked:', n_checked, file=sys.stderr)
    print('bp consumed:', bp_consumed, bp_consumed / float(C), file=sys.stderr)
    print('error rate: %.2f%%' % \
        (100.0 * sum(positions) / float(sum(lengths))), file=sys.stderr)

    print('Error histogram is in %s' % output_filename, file=sys.stderr)

    if not exit_condition(n_consumed, n_checked):
        print("", file=sys.stderr)
        print("** WARNING: not enough reads to get a good result", file=sys.stderr)
        print("** Is this high diversity sample / small subset?", file=sys.stderr)
        sys.exit(-1)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2010-2015, Michigan State University.
# Copyright (C) 2015-2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=invalid-name,missing-docstring
"""
Find highly-connected k-mers.

k-mers are output into a .stoptags file, for later use in partitioning.

% python scripts/find-knots.py <base>
"""
from __future__ import print_function

import glob
import os
import textwrap
import khmer
import sys
from khmer.kfile import check_input_files, check_space
from khmer import khmer_args
from khmer.khmer_args import (build_counting_args, sanitize_help)

# counting hash parameters.
DEFAULT_COUNTING_HT_SIZE = 3e6                # number of bytes
DEFAULT_COUNTING_HT_N = 4                     # number of counting hash tables

# Lump removal parameters.  Probably shouldn't be changed, but who knows?
#
# explanation:
#
# We will walk EXCURSION_DISTANCE out from each tag; if we find more than
# EXCURSION_KMER_THRESHOLD kmers within that range, this will be a "big"
# excursion and we will track all k-mers visited.  If we find that any
# k-mer has been visited more than EXCURSION_KMER_COUNT_THRESHOLD times,
# we will mark it as BAD and make it a stop tag for traversal.

# don't change these!
EXCURSION_DISTANCE = 40
EXCURSION_KMER_THRESHOLD = 200
EXCURSION_KMER_COUNT_THRESHOLD = 2
# EXCURSION_KMER_COUNT_THRESHOLD=5 # -- works ok for non-diginormed data


def get_parser():
    epilog = """\
    Load an k-mer nodegraph/tagset pair created by
    :program:`load-graph.py`, and a set of pmap files created by
    :program:`partition-graph.py`. Go through each pmap file,
    select the largest partition in each, and do the same kind of traversal as
    in :program:`make-initial-stoptags.py` from each of the waypoints in that
    partition; this should identify all of the Highly Connected Kmers in that
    partition. These HCKs are output to ``<graphbase>.stoptags`` after each
    pmap file.

    Parameter choice is reasonably important. See the pipeline in
    :doc:`partitioning-big-data` for an example run.

    This script is not very scalable and may blow up memory and die horribly.
    You should be able to use the intermediate stoptags to restart the
    process, and if you eliminate the already-processed pmap files, you can
    continue where you left off.
    """
    parser = build_counting_args(
        descr="Find all highly connected k-mers.",
        epilog=textwrap.dedent(epilog),
        citations=['graph'])

    parser.add_argument('graphbase', help='Basename for the input and output '
                        'files.')
    parser.add_argument('-f', '--force', default=False, action='store_true',
                        help='Continue past warnings')
    return parser


def main():
    parser = get_parser()
    parser.epilog = parser.epilog.replace(
        ":doc:`partitioning-big-data`",
        "http://khmer.readthedocs.io/en/stable/user/"
        "partitioning-big-data.html"
    )
    args = sanitize_help(parser).parse_args()

    graphbase = args.graphbase

    # @RamRS: This might need some more work
    infiles = [graphbase, graphbase + '.tagset']
    if os.path.exists(graphbase + '.stoptags'):
        infiles.append(graphbase + '.stoptags')
    for _ in infiles:
        check_input_files(_, args.force)

    check_space(infiles, args.force)

    print('loading k-mer nodegraph %s' % graphbase, file=sys.stderr)
    graph = khmer.load_nodegraph(graphbase)

    print('loading tagset %s.tagset...' % graphbase, file=sys.stderr)
    graph.load_tagset(graphbase + '.tagset')

    initial_stoptags = False    # @CTB regularize with make-initial
    if os.path.exists(graphbase + '.stoptags'):
        print('loading stoptags %s.stoptags' % graphbase, file=sys.stderr)
        graph.load_stop_tags(graphbase + '.stoptags')
        initial_stoptags = True

    pmap_files = glob.glob(args.graphbase + '.subset.*.pmap')

    print('loading %d pmap files (first one: %s)' %
          (len(pmap_files), pmap_files[0]), file=sys.stderr)
    print('---', file=sys.stderr)
    print('output stoptags will be in',
          graphbase + '.stoptags', file=sys.stderr)
    if initial_stoptags:
        print(
            '(these output stoptags will include the already-loaded set)',
            file=sys.stderr)
    print('---', file=sys.stderr)

    # create countgraph
    ksize = graph.ksize()
    counting = khmer_args.create_countgraph(args, ksize=ksize)

    # load & merge
    for index, subset_file in enumerate(pmap_files):
        print('<-', subset_file, file=sys.stderr)
        subset = graph.load_subset_partitionmap(subset_file)

        print('** repartitioning subset... %s' % subset_file, file=sys.stderr)
        graph.repartition_largest_partition(subset, counting,
                                            EXCURSION_DISTANCE,
                                            EXCURSION_KMER_THRESHOLD,
                                            EXCURSION_KMER_COUNT_THRESHOLD)

        print('** merging subset... %s' % subset_file, file=sys.stderr)
        graph.merge_subset(subset)

        print('** repartitioning, round 2... %s' %
              subset_file, file=sys.stderr)
        size = graph.repartition_largest_partition(
            None, counting, EXCURSION_DISTANCE, EXCURSION_KMER_THRESHOLD,
            EXCURSION_KMER_COUNT_THRESHOLD)

        print('** repartitioned size:', size, file=sys.stderr)

        print('saving stoptags binary', file=sys.stderr)
        graph.save_stop_tags(graphbase + '.stoptags')
        os.rename(subset_file, subset_file + '.processed')
        print('(%d of %d)\n' % (index, len(pmap_files)), file=sys.stderr)

    print('done!', file=sys.stderr)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2011-2015, Michigan State University.
# Copyright (C) 2015, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the Michigan State University nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
from __future__ import print_function
import sys
import os
import khmer
from khmer.thread_utils import ThreadedSequenceProcessor, verbose_fasta_iter

WORKER_THREADS = 8
GROUPSIZE = 100

CUTOFF = 50

###


def main():
    counting_ht = sys.argv[1]
    infiles = sys.argv[2:]

    print('file with ht: %s' % counting_ht)
    print('-- settings:')
    print('N THREADS', WORKER_THREADS)
    print('--')

    print('making hashtable')
    ht = khmer.load_countgraph(counting_ht)
    K = ht.ksize()

    for infile in infiles:
        print('filtering', infile)
        outfile = os.path.basename(infile) + '.below'

        outfp = open(outfile, 'w')

        def process_fn(record, ht=ht):
            name = record['name']
            seq = record['sequence']
            if 'N' in seq:
                return None, None

            trim_seq, trim_at = ht.trim_below_abundance(seq, CUTOFF)

            if trim_at >= K:
                return name, trim_seq

            return None, None

        tsp = ThreadedSequenceProcessor(process_fn, WORKER_THREADS, GROUPSIZE)

        tsp.start(verbose_fasta_iter(infile), outfp)

if __name__ == '__main__':
    main()



# This file is part of khmer, https://github.com/dib-lab/khmer/, and is
# Copyright (C) 2016, The Regents of the University of California.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * Neither the name of the University of California nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Contact: khmer-project@idyll.org
# pylint: disable=C0111,C0103,missing-docstring,no-member,protected-access

from __future__ import print_function
from __future__ import absolute_import

import khmer

import pytest
from . import khmer_tst_utils as utils


def test_get_kmer_hashes():
    s = "ATGGATATGGAGGACAAGTATATGGAGGACAAGTATATGGAGGACAAGTAT"
    a = khmer.Counttable(33, 1e6, 3)
    assert a.get_kmer_hashes(s[:33]) == [4743239192574154715]
    assert a.get_kmer_hashes(s[:34]) == [4743239192574154715,
                                         2122462908541313313]

    assert a.get_kmer_hashes(s[0:33]) == [4743239192574154715]
    assert a.get_kmer_hashes(s[1:34]) == [2122462908541313313]


@pytest.mark.parametrize('kmer', [
    ('GATTACA' * 3),
    ('ATG' * 7),
    ('AGGACAAGTATATGGAGGACA'),
])
def test_kmer_revcom_hash(kmer):
    a = khmer.Counttable(21, 1e4, 3)
    assert a.hash(kmer) == a.hash(khmer.reverse_complement(kmer))


@pytest.mark.parametrize('ksize,sketch_allocator', [
    (21, khmer.Nodetable),
    (21, khmer.Counttable),
    (21, khmer.SmallCounttable),
    (49, khmer.Nodetable),
    (49, khmer.Counttable),
    (49, khmer.SmallCounttable),
])
def test_reverse_hash(ksize, sketch_allocator):
    multiplier = int(ksize / len('GATTACA'))
    kmer = 'GATTACA' * multiplier

    sketch = sketch_allocator(ksize, 1e4, 4)
    kmer_hash = sketch.hash(kmer)
    with pytest.raises(ValueError) as ve:
        _ = sketch.reverse_hash(kmer_hash)
    assert 'not implemented' in str(ve)
